@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   bloom.py                                                                  #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -20,6 +20,7 @@ import abc
 import functools
 import itertools
 import math
+import uuid
 from typing import Any
 from typing import Callable
 from typing import Generator
@@ -28,12 +29,18 @@ from typing import Set
 from typing import cast
 
 import mmh3
+# TODO: When we drop support for Python 3.7, change the following import to:
+#   from typing import final
+from typing_extensions import final
 
 from .annotations import F
-from .base import Base
-from .base import JSONTypes
+from .annotations import JSONTypes
+from .base import Container
 
 
+# TODO: When we drop support for Python 3.7, stop using @_store_on_self().  Use
+# @functools.cached_property() instead.
+#   https://docs.python.org/3/library/functools.html#functools.cached_property
 def _store_on_self(*, attr: str) -> Callable[[F], F]:
     "Decorator to store/cache a method's return value as an attribute on self."
     def decorator(func: F) -> F:
@@ -49,7 +56,7 @@ def _store_on_self(*, attr: str) -> Callable[[F], F]:
     return decorator
 
 
-class BloomFilterABC(metaclass=abc.ABCMeta):
+class BloomFilterABC(abc.ABC):
     '''Bloom filter abstract base class.
 
     This abstract base class:
@@ -60,25 +67,24 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
     '''
 
     @abc.abstractmethod
-    def _bit_offsets(self,
-                     encoded_value: JSONTypes,
-                     ) -> Generator[int, None, None]:
+    def _bit_offsets(self, value: JSONTypes) -> Generator[int, None, None]:
         for seed in range(self.num_hashes()):
-            hash_ = mmh3.hash(encoded_value, seed=seed)
+            hash_ = mmh3.hash(cast(str, value), seed=seed)
             yield hash_ % self.size()
 
+    @abc.abstractmethod
+    def _num_bits_set(self) -> int:
+        raise NotImplementedError
+
     @abc.abstractmethod
     def update(self, *iterables: Iterable[JSONTypes]) -> None:
         raise NotImplementedError
 
     @abc.abstractmethod
-    def __contains__(self, value: JSONTypes) -> bool:
-        raise NotImplementedError
-
-    @abc.abstractmethod
-    def _num_bits_set(self) -> int:
+    def contains_many(self, *values: JSONTypes) -> Generator[bool, None, None]:
         raise NotImplementedError
 
+    @final
     def __init__(self,
                  iterable: Iterable[JSONTypes] = frozenset(),
                  *args: Any,
@@ -92,11 +98,12 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         into this Bloom filter, and k is the number of times to run our hash
         functions on each element.
         '''
-        super().__init__(*args, **kwargs)  # type: ignore
+        super().__init__(*args, **kwargs)
         self.num_elements = num_elements
         self.false_positives = false_positives
         self.update(iterable)
 
+    @final
     @_store_on_self(attr='_size')
     def size(self) -> int:
         '''The required number of bits (m) given n and p.
@@ -117,6 +124,7 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         size = math.ceil(size)
         return size
 
+    @final
     @_store_on_self(attr='_num_hashes')
     def num_hashes(self) -> int:
         '''The number of hash functions (k) given m and n, minimizing p.
@@ -134,6 +142,20 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         num_hashes = math.ceil(num_hashes)
         return num_hashes
 
+    @final
+    def _bit_offsets_many(self,
+                          *values: JSONTypes,
+                          ) -> Generator[int, None, None]:
+        for value in values:
+            try:
+                yield from self._bit_offsets(value)
+            except TypeError:
+                # value can't be encoded / converted to JSON.  Do a membership
+                # test for a UUID in place of value.
+                uuid4 = str(uuid.uuid4())
+                yield from self._bit_offsets(uuid4)
+
+    @final
     def add(self, value: JSONTypes) -> None:
         '''Add an element to the BloomFilter.  O(k)
 
@@ -143,6 +165,7 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         '''
         self.update({value})
 
+    @final
     def __len__(self) -> int:
         '''Return the approximate the number of elements in the BloomFilter.  O(m)
 
@@ -150,7 +173,9 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         Bloom filter.
 
         Please note that this method returns an approximation, not an exact
-        value, though it's quite accurate.
+        value, though it's quite accurate.  The accuracy can be tuned with the
+        num_elements and false_positives arguments to .__init__(), at the
+        expense of storage and CPU time.
 
         More about the formula that this method implements:
             https://en.wikipedia.org/wiki/Bloom_filter#Approximating_the_number_of_items_in_a_Bloom_filter
@@ -162,8 +187,24 @@ class BloomFilterABC(metaclass=abc.ABCMeta):
         )
         return math.floor(len_)
 
+    @final
+    def __contains__(self, value: JSONTypes) -> bool:
+        '''bf.__contains__(element) <==> element in bf.  O(k)
 
-class BloomFilter(BloomFilterABC, Base):
+        Here, k is the number of times to run our hash functions on a given
+        input string to compute bit offests into the underlying string
+        representing this Bloom filter.
+
+        Please note that this method *may* return false positives, but *never*
+        returns false negatives.  This means that if `element in bf` evaluates
+        to True, then you *may* have inserted the element into the Bloom filter.
+        But if `element in bf` evaluates to False, then you *must not* have
+        inserted it.
+        '''
+        return next(self.contains_many(value))
+
+
+class BloomFilter(BloomFilterABC, Container):
     '''Redis-backed Bloom filter with an API similar to Python sets.
 
     Bloom filters are a powerful data structure that help you to answer the
@@ -208,7 +249,7 @@ class BloomFilter(BloomFilterABC, Base):
 
         >>> dilberts.add('rajiv')
 
-    Test for membership in the Bloom filter:
+    Test for element membership in the Bloom filter:
 
         >>> 'rajiv' in dilberts
         True
@@ -228,6 +269,8 @@ class BloomFilter(BloomFilterABC, Base):
     Insert multiple elements into the Bloom filter:
 
         >>> dilberts.update({'raj', 'dan'})
+        >>> len(dilberts)
+        3
 
     Remove all of the elements from the Bloom filter:
 
@@ -268,6 +311,14 @@ class BloomFilter(BloomFilterABC, Base):
         encoded_value = self._encode(value)
         return super()._bit_offsets(encoded_value)
 
+    def _num_bits_set(self) -> int:
+        '''The number of bits set to 1 in this Bloom filter.  O(m)
+
+        Here, m is the size in bits of the underlying string representing this
+        Bloom filter.
+        '''
+        return self.redis.bitcount(self.key)  # Available since Redis 2.6.0
+
     def update(self, *iterables: Iterable[JSONTypes]) -> None:
         '''Populate the Bloom filter with the elements in iterables.  O(n * k)
 
@@ -280,31 +331,37 @@ class BloomFilter(BloomFilterABC, Base):
             for value in itertools.chain(*iterables):
                 bit_offsets.update(self._bit_offsets(value))
 
-            pipeline.multi()
+            pipeline.multi()  # Available since Redis 1.2.0
             for bit_offset in bit_offsets:
-                pipeline.setbit(self.key, bit_offset, 1)
+                pipeline.setbit(self.key, bit_offset, 1)  # Available since Redis 2.2.0
 
-    def __contains__(self, value: JSONTypes) -> bool:
-        '''bf.__contains__(element) <==> element in bf.  O(k)
+    def contains_many(self, *values: JSONTypes) -> Generator[bool, None, None]:
+        '''Yield whether this Bloom filter contains multiple elements.  O(n)
 
-        Here, k is the number of times to run our hash functions on a given
-        input string to compute bit offests into the underlying string
-        representing this Bloom filter.
+        Please note that this method *may* produce false positives, but *never*
+        produces false negatives.  This means that if .contains_many() yields
+        True, then you *may* have inserted the element into the Bloom filter.
+        But if .contains_many() yields False, then you *must not* have inserted
+        it.
         '''
         with self._watch() as pipeline:
-            pipeline.multi()
-            for bit_offset in self._bit_offsets(value):
-                pipeline.getbit(self.key, bit_offset)
-            bits = pipeline.execute()
-        return all(bits)
+            pipeline.multi()  # Available since Redis 1.2.0
+            for bit_offset in self._bit_offsets_many(*values):
+                pipeline.getbit(self.key, bit_offset)  # Available since Redis 2.2.0
+            bits = iter(pipeline.execute())  # Available since Redis 1.2.0
 
-    def _num_bits_set(self) -> int:
-        '''The number of bits set to 1 in this Bloom filter.  O(m)
-
-        Here, m is the size in bits of the underlying string representing this
-        Bloom filter.
-        '''
-        return self.redis.bitcount(self.key)
+        # I stole this recipe from here:
+        #   https://stackoverflow.com/a/61435714
+        #
+        # TODO: When we drop support for Python 3.7, rewrite the following loop
+        # using the walrus operator, like in the Stack Overflow answer linked
+        # above.
+        bits_per_chunk = self.num_hashes()
+        while True:
+            bits_in_chunk = tuple(itertools.islice(bits, bits_per_chunk))
+            if not bits_in_chunk:
+                break
+            yield all(bits_in_chunk)
 
     def __repr__(self) -> str:
         'Return the string representation of the BloomFilter.  O(1)'
@@ -318,5 +375,5 @@ if __name__ == '__main__':
     #   $ deactivate
     import contextlib
     with contextlib.suppress(ImportError):
-        from tests.base import run_doctests  # type: ignore
+        from tests.base import run_doctests
         run_doctests()
