@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   cache.py                                                                  #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -16,39 +16,47 @@
 # --------------------------------------------------------------------------- #
 
 
+# TODO: When we drop support for Python 3.9, remove the following import.  We
+# only need it for X | Y union type annotations as of 2022-01-29.
+from __future__ import annotations
+
 import collections
-import contextlib
 import functools
 import itertools
-import logging
 from typing import Any
 from typing import Callable
 from typing import ClassVar
 from typing import Collection
 from typing import Hashable
 from typing import Iterable
+from typing import Mapping
 from typing import NamedTuple
-from typing import Optional
+from typing import Tuple
 from typing import TypeVar
 from typing import Union
 from typing import cast
 
 from redis import Redis
 from redis.exceptions import WatchError
+# TODO: When we drop support for Python 3.7, change the following import to:
+#   from typing import Final
 from typing_extensions import Final
 
-from .base import JSONTypes
+from .annotations import JSONTypes
 from .base import _default_redis
+from .base import logger
 from .base import random_key
-from .dict import InitArg
-from .dict import InitIter
-from .dict import InitMap
 from .dict import RedisDict
 
 
-_DEFAULT_TIMEOUT: Final[int] = 60   # seconds
+F = TypeVar('F', bound=Callable[..., JSONTypes])
 
-_logger: Final[logging.Logger] = logging.getLogger('pottery')
+UpdateMap = Mapping[JSONTypes, Union[JSONTypes, object]]
+UpdateItem = Tuple[JSONTypes, Union[JSONTypes, object]]
+UpdateIter = Iterable[UpdateItem]
+UpdateArg = Union[UpdateMap, UpdateIter]
+
+_DEFAULT_TIMEOUT: Final[int] = 60   # seconds
 
 
 class CacheInfo(NamedTuple):
@@ -59,7 +67,7 @@ class CacheInfo(NamedTuple):
     '''
     hits: int = 0
     misses: int = 0
-    maxsize: Optional[int] = None
+    maxsize: int | None = None
     currsize: int = 0
 
 
@@ -68,13 +76,10 @@ def _arg_hash(*args: Hashable, **kwargs: Hashable) -> int:
     return hash((args, kwargs_items))
 
 
-F = TypeVar('F', bound=Callable[..., JSONTypes])
-
-
 def redis_cache(*,  # NoQA: C901
-                redis: Optional[Redis] = None,
-                key: Optional[str] = None,
-                timeout: Optional[int] = _DEFAULT_TIMEOUT,
+                redis: Redis | None = None,
+                key: str | None = None,
+                timeout: int | None = _DEFAULT_TIMEOUT,
                 ) -> Callable[[F], F]:
     '''Redis-backed caching decorator with an API like functools.lru_cache().
 
@@ -123,13 +128,14 @@ def redis_cache(*,  # NoQA: C901
     on each call.
     '''
 
-    redis = _default_redis if redis is None else redis
+    if redis is None:
+        redis = _default_redis
 
     def decorator(func: F) -> F:
         nonlocal redis, key
-        if key is None:  # pragma: no cover
+        if key is None:
             key = random_key(redis=cast(Redis, redis))
-            _logger.warning(
+            logger.warning(
                 "Self-assigning key redis_cache(key='%s') for function %s",
                 key,
                 func.__qualname__,
@@ -171,7 +177,7 @@ def redis_cache(*,  # NoQA: C901
 
         def cache_clear() -> None:
             nonlocal hits, misses
-            cast(Redis, redis).delete(cast(str, key))
+            cast(Redis, redis).unlink(cast(str, key))
             hits, misses = 0, 0
 
         wrapper.__wrapped__ = func  # type: ignore
@@ -184,10 +190,10 @@ def redis_cache(*,  # NoQA: C901
 
 def _set_expiration(func: F) -> F:
     @functools.wraps(func)
-    def wrapper(self, *args: Any, **kwargs: Any) -> Any:
+    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
         value = func(self, *args, **kwargs)
         if self._timeout:
-            self._cache.redis.expire(self._cache.key, self._timeout)
+            self._cache.redis.expire(self._cache.key, self._timeout)  # Available since Redis 1.0.0
         return value
     return cast(F, wrapper)
 
@@ -217,11 +223,11 @@ class CachedOrderedDict(collections.OrderedDict):
     @_set_expiration
     def __init__(self,
                  *,
-                 redis_client: Optional[Redis] = None,
-                 redis_key: Optional[str] = None,
+                 redis_client: Redis | None = None,
+                 redis_key: str | None = None,
                  dict_keys: Iterable[JSONTypes] = tuple(),
                  num_tries: int = _NUM_TRIES,
-                 timeout: Optional[int] = _DEFAULT_TIMEOUT,
+                 timeout: int | None = _DEFAULT_TIMEOUT,
                  ) -> None:
         self._num_tries = num_tries
         self._timeout = timeout
@@ -241,19 +247,20 @@ class CachedOrderedDict(collections.OrderedDict):
             encoded_keys = (
                 self._cache._encode(dict_key) for dict_key in dict_keys
             )
-            encoded_values = self._cache.redis.hmget(
+            encoded_values = self._cache.redis.hmget(  # Available since Redis 2.0.0
                 self._cache.key,
                 *encoded_keys,
             )
             for dict_key, encoded_value in zip(dict_keys, encoded_values):
                 if encoded_value is None:
-                    value = self._SENTINEL
                     self._misses.add(dict_key)
+                    value = self._SENTINEL
                 else:
                     value = self._cache._decode(encoded_value)
                 item = (dict_key, value)
                 items.append(item)
-        return super().__init__(items)
+        super().__init__()
+        self.__update(items)
 
     def misses(self) -> Collection[JSONTypes]:
         return frozenset(self._misses)
@@ -261,13 +268,13 @@ class CachedOrderedDict(collections.OrderedDict):
     @_set_expiration
     def __setitem__(self,
                     dict_key: JSONTypes,
-                    value: Union[JSONTypes, object]
+                    value: JSONTypes | object,
                     ) -> None:
         'Set self[dict_key] to value.'
-        if value is not self._SENTINEL:
+        if value is not self._SENTINEL:  # pragma: no cover
             self._cache[dict_key] = value
             self._misses.discard(dict_key)
-        return super().__setitem__(dict_key, value)
+        super().__setitem__(dict_key, value)
 
     @_set_expiration
     def setdefault(self,
@@ -300,15 +307,15 @@ class CachedOrderedDict(collections.OrderedDict):
                                ) -> None:
         with self._cache._watch() as pipeline:
             if dict_key not in self._cache:
-                pipeline.multi()
+                pipeline.multi()  # Available since Redis 1.2.0
                 # The following line is equivalent to: self._cache[dict_key] = default
-                pipeline.hset(
+                pipeline.hset(  # Available since Redis 2.0.0
                     self._cache.key,
                     self._cache._encode(dict_key),
                     self._cache._encode(default),
                 )
 
-    def __retry(self, callable: Callable[[], Any], try_num: int = 0) -> Any:
+    def __retry(self, callable: Callable[[], Any], *, try_num: int = 0) -> Any:
         try:
             return callable()
         except WatchError:  # pragma: no cover
@@ -317,23 +324,29 @@ class CachedOrderedDict(collections.OrderedDict):
             raise
 
     @_set_expiration
-    def update(self, arg: InitArg = tuple(), **kwargs: JSONTypes) -> None:  # type: ignore
+    def update(self,  # type: ignore
+               arg: UpdateArg = tuple(),
+               **kwargs: JSONTypes | object,
+               ) -> None:
         '''D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.
         If E is present and has an .items() method, then does:  for k in E: D[k] = E[k]
         If E is present and lacks an .items() method, then does:  for k, v in E: D[k] = v
         In either case, this is followed by: for k in F:  D[k] = F[k]
 
         The base class, OrderedDict, has an .update() method that works just
-        fine.  The trouble is that it executes multiple calls to .__setitem__()
-        therefore multiple round trips to Redis.  This overridden .update()
-        makes a single bulk call to Redis.
+        fine.  The trouble is that it executes multiple calls to
+        self.__setitem__() therefore multiple round trips to Redis.  This
+        overridden .update() makes a single bulk call to Redis.
         '''
         to_cache = {}
-        with contextlib.suppress(AttributeError):
-            arg = cast(InitMap, arg).items()
-        for dict_key, value in itertools.chain(cast(InitIter, arg), kwargs.items()):
+        if isinstance(arg, collections.abc.Mapping):
+            arg = arg.items()
+        items = itertools.chain(arg, kwargs.items())
+        for dict_key, value in items:
             if value is not self._SENTINEL:
                 to_cache[dict_key] = value
-                self._misses.discard(dict_key)
+                self._misses.discard(cast(JSONTypes, dict_key))
             super().__setitem__(dict_key, value)
         self._cache.update(to_cache)
+
+    __update = update
