@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   counter.py                                                                #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -16,13 +16,19 @@
 # --------------------------------------------------------------------------- #
 
 
+# TODO: Remove the following import after deferred evaluation of annotations
+# because the default.
+#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
+#   2. https://www.python.org/dev/peps/pep-0563/
+#   3. https://www.python.org/dev/peps/pep-0649/
+from __future__ import annotations
+
 import collections
 import contextlib
 import itertools
 from typing import Callable
 from typing import Iterable
 from typing import List
-from typing import Optional
 from typing import Tuple
 from typing import Union
 from typing import cast
@@ -30,7 +36,7 @@ from typing import cast
 from redis.client import Pipeline
 from typing_extensions import Counter
 
-from .base import JSONTypes
+from .annotations import JSONTypes
 from .dict import RedisDict
 
 
@@ -50,23 +56,29 @@ class RedisCounter(RedisDict, collections.Counter):
                   sign: int = +1,
                   **kwargs: int,
                   ) -> None:
-        to_set = {}
-        try:
-            for key, value in cast(Counter[JSONTypes], arg).items():
-                to_set[key] = sign * value
-        except AttributeError:
+        dict_ = {}
+        if isinstance(arg, collections.abc.Mapping):
+            items = arg.items()
+            for key, value in items:
+                dict_[key] = sign * value
+        else:
             for key in arg:
-                to_set[key] = to_set.get(key, self[key]) + sign
+                value = dict_.get(key, self[key])
+                dict_[key] = value + sign
+
         for key, value in kwargs.items():
-            original = self[key] if to_set.get(key, 0) == 0 else to_set[key]
-            to_set[key] = original + sign * value
-        to_set = {key: self[key] + value for key, value in to_set.items()}
-        encoded_to_set = {
-            self._encode(k): self._encode(v) for k, v in to_set.items()
-        }
-        if encoded_to_set:
-            pipeline.multi()
-            pipeline.hset(self.key, mapping=encoded_to_set)  # type: ignore
+            if dict_.get(key, 0) == 0:
+                original = self[key]
+            else:  # pragma: no cover
+                original = dict_[key]
+            dict_[key] = original + sign * value
+
+        dict_ = {key: self[key] + value for key, value in dict_.items()}
+        encoded_dict = self._encode_dict(dict_)
+        if encoded_dict:
+            pipeline.multi()  # Available since Redis 1.2.0
+            # Available since Redis 2.0.0:
+            pipeline.hset(self.key, mapping=encoded_dict)  # type: ignore
 
     # Preserve the Open-Closed Principle with name mangling.
     #   https://youtu.be/miGolgp9xq8?t=2086
@@ -98,26 +110,16 @@ class RedisCounter(RedisDict, collections.Counter):
 
     def __repr__(self) -> str:
         'Return the string representation of the RedisCounter.  O(n)'
-        items = self.most_common()
+        items = self.__most_common()
         pairs = (f"'{key}': {value}" for key, value in items)
         repr_ = ', '.join(pairs)
         return self.__class__.__name__ + '{' + repr_ + '}'
 
-    def _make_counter(self) -> Counter[JSONTypes]:
-        counter: Counter[JSONTypes] = collections.Counter()
-        cursor = 0
-        while True:
-            cursor, encoded_dict = self._scan(cursor=cursor)
-            decoded_dict = {
-                self._decode(key): self._decode(value)
-                for key, value in encoded_dict.items()
-            }
-            counter.update(decoded_dict)
-            if cursor == 0:
-                break
-        return counter
+    def to_counter(self) -> Counter[JSONTypes]:
+        'Convert a RedisCounter into a plain Python collections.Counter.'
+        return collections.Counter(self)
 
-    __make_counter = _make_counter
+    __to_counter = to_counter
 
     def __math_op(self,
                   other: Counter[JSONTypes],
@@ -125,7 +127,7 @@ class RedisCounter(RedisDict, collections.Counter):
                   method: Callable[[Counter[JSONTypes], Counter[JSONTypes]], Counter[JSONTypes]],
                   ) -> Counter[JSONTypes]:
         with self._watch(other):
-            counter = self.__make_counter()
+            counter = self.__to_counter()
             return method(counter, other)
 
     def __add__(self, other: Counter[JSONTypes]) -> Counter[JSONTypes]:
@@ -151,7 +153,7 @@ class RedisCounter(RedisDict, collections.Counter):
                    ) -> Counter[JSONTypes]:
         with self._watch():
             counter: Counter[JSONTypes] = collections.Counter()
-            for key, value in self.__make_counter().items():
+            for key, value in self.__to_counter().items():
                 if test_func(value):
                     counter[key] = modifier_func(value)
             return counter
@@ -174,10 +176,10 @@ class RedisCounter(RedisDict, collections.Counter):
                    other: Counter[JSONTypes],
                    *,
                    sign: int = +1,
-                   ) -> 'RedisCounter':
+                   ) -> RedisCounter:
         with self._watch(other) as pipeline:
             try:
-                other_items = cast('RedisCounter', other)._make_counter().items()
+                other_items = cast(RedisCounter, other).to_counter().items()
             except AttributeError:
                 other_items = other.items()
             to_set = {k: self[k] + sign * v for k, v in other_items}
@@ -190,11 +192,12 @@ class RedisCounter(RedisDict, collections.Counter):
             }
             encoded_to_del = {self._encode(k) for k in to_del}
             if encoded_to_set or encoded_to_del:
-                pipeline.multi()
+                pipeline.multi()  # Available since Redis 1.2.0
                 if encoded_to_set:
+                    # Available since Redis 2.0.0:
                     pipeline.hset(self.key, mapping=encoded_to_set)  # type: ignore
                 if encoded_to_del:
-                    pipeline.hdel(self.key, *encoded_to_del)
+                    pipeline.hdel(self.key, *encoded_to_del)  # Available since Redis 2.0.0
             return self
 
     def __iadd__(self, other: Counter[JSONTypes]) -> Counter[JSONTypes]:
@@ -209,11 +212,11 @@ class RedisCounter(RedisDict, collections.Counter):
                   other: Counter[JSONTypes],
                   *,
                   method: Callable[[int, int], bool],
-                  ) -> 'RedisCounter':
+                  ) -> RedisCounter:
         with self._watch(other) as pipeline:
-            self_counter = self.__make_counter()
+            self_counter = self.__to_counter()
             try:
-                other_counter = cast('RedisCounter', other)._make_counter()
+                other_counter = cast(RedisCounter, other).to_counter()
             except AttributeError:
                 other_counter = other
             to_set, to_del = {}, set()
@@ -226,17 +229,17 @@ class RedisCounter(RedisDict, collections.Counter):
                     del to_set[k]
                     to_del.add(k)
             if to_set or to_del:
-                pipeline.multi()
-                if to_set:
-                    encoded_to_set = {
-                        self._encode(k): self._encode(v)
-                        for k, v in to_set.items()
-                    }
-                    pipeline.hset(self.key, mapping=encoded_to_set)  # type: ignore
-                if to_del:
-                    encoded_to_del = {self._encode(k) for k in to_del}
-                    pipeline.hdel(self.key, *encoded_to_del)
-            return self
+                pipeline.multi()  # Available since Redis 1.2.0
+            if to_set:
+                encoded_to_set = {
+                    self._encode(k): self._encode(v) for k, v in to_set.items()
+                }
+                # Available since Redis 2.0.0:
+                pipeline.hset(self.key, mapping=encoded_to_set)  # type: ignore
+            if to_del:
+                encoded_to_del = {self._encode(k) for k in to_del}
+                pipeline.hdel(self.key, *encoded_to_del)  # Available since Redis 2.0.0
+        return self
 
     def __ior__(self, other: Counter[JSONTypes]) -> Counter[JSONTypes]:  # type: ignore
         'Same as __or__(), but in-place.  O(n)'
@@ -247,7 +250,9 @@ class RedisCounter(RedisDict, collections.Counter):
         return self.__iset_op(other, method=int.__lt__)
 
     def most_common(self,
-                    n: Optional[int] = None,
+                    n: int | None = None,
                     ) -> List[Tuple[JSONTypes, int]]:
-        counter = self.__make_counter()
+        counter = self.__to_counter()
         return counter.most_common(n=n)
+
+    __most_common = most_common
