@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   deque.py                                                                  #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -16,31 +16,37 @@
 # --------------------------------------------------------------------------- #
 
 
+# TODO: When we drop support for Python 3.9, remove the following import.  We
+# only need it for X | Y union type annotations as of 2022-01-29.
+from __future__ import annotations
+
 import collections
+import warnings
 from typing import Iterable
-from typing import Optional
 from typing import Tuple
-from typing import Union
 from typing import cast
 
 from redis import Redis
 from redis.client import Pipeline
 
-from .base import JSONTypes
+from .annotations import JSONTypes
+from .exceptions import InefficientAccessWarning
 from .list import RedisList
 
 
 class RedisDeque(RedisList, collections.deque):  # type: ignore
     'Redis-backed container compatible with collections.deque.'
 
-    # Method overrides:
+    # Overrides:
+
+    _ALLOWED_TO_EQUAL = collections.deque
 
     def __init__(self,
                  iterable: Iterable[JSONTypes] = tuple(),
-                 maxlen: Optional[int] = None,
+                 maxlen: int | None = None,
                  *,
-                 redis: Optional[Redis] = None,
-                 key: Optional[str] = None,
+                 redis: Redis | None = None,
+                 key: str = '',
                  ) -> None:
         'Initialize the RedisDeque.  O(n)'
         if maxlen is not None and not isinstance(maxlen, int):
@@ -66,7 +72,7 @@ class RedisDeque(RedisList, collections.deque):  # type: ignore
         super()._populate(pipeline, iterable)
 
     @property
-    def maxlen(self) -> Optional[int]:
+    def maxlen(self) -> int | None:
         return self._maxlen
 
     @maxlen.setter
@@ -79,7 +85,7 @@ class RedisDeque(RedisList, collections.deque):  # type: ignore
     def insert(self, index: int, value: JSONTypes) -> None:
         'Insert an element into the RedisDeque before the given index.  O(n)'
         with self._watch() as pipeline:
-            current_length = cast(int, pipeline.llen(self.key))
+            current_length = cast(int, pipeline.llen(self.key))  # Available since Redis 1.0.0
             if self.maxlen is not None and current_length >= self.maxlen:
                 raise IndexError(
                     f'{self.__class__.__name__} already at its maximum size'
@@ -118,16 +124,16 @@ class RedisDeque(RedisList, collections.deque):  # type: ignore
         with self._watch(values) as pipeline:
             push_method_name = 'rpush' if right else 'lpush'
             encoded_values = [self._encode(value) for value in values]
-            len_ = cast(int, pipeline.llen(self.key)) + len(encoded_values)
-            trim_indices: Union[Tuple[int, int], Tuple] = tuple()
+            len_ = cast(int, pipeline.llen(self.key)) + len(encoded_values)  # Available since Redis 1.0.0
+            trim_indices: Tuple[int, int] | Tuple = tuple()
             if self.maxlen is not None and len_ >= self.maxlen:
                 trim_indices = (len_-self.maxlen, len_) if right else (0, self.maxlen-1)
 
-            pipeline.multi()
+            pipeline.multi()  # Available since Redis 1.2.0
             push_method = getattr(pipeline, push_method_name)
             push_method(self.key, *encoded_values)
             if trim_indices:
-                pipeline.ltrim(self.key, *trim_indices)
+                pipeline.ltrim(self.key, *trim_indices)  # Available since Redis 1.0.0
 
     def pop(self) -> JSONTypes:  # type: ignore
         return super().pop()
@@ -154,15 +160,15 @@ class RedisDeque(RedisList, collections.deque):  # type: ignore
                 # Rotating an empty RedisDeque is a no-op.
                 return
 
-            push_method_name = 'lpush' if n > 0 else 'rpush'
-            decoded_values = self[-n:][::-1] if n > 0 else self[:-n]
-            encoded_values = (self._encode(value) for value in decoded_values)
+            push_method_name = 'lpush' if n > 0 else 'rpush'  # Available since Redis 1.0.0
+            values = self[-n:][::-1] if n > 0 else self[:-n]
+            encoded_values = (self._encode(value) for value in values)
             trim_indices = (0, len(self)-1) if n > 0 else (-n, len(self)-1-n)
 
-            pipeline.multi()
+            pipeline.multi()  # Available since Redis 1.2.0
             push_method = getattr(pipeline, push_method_name)
             push_method(self.key, *encoded_values)
-            pipeline.ltrim(self.key, *trim_indices)
+            pipeline.ltrim(self.key, *trim_indices)  # Available since Redis 1.0.0
 
     # Methods required for Raj's sanity:
 
@@ -172,8 +178,12 @@ class RedisDeque(RedisList, collections.deque):  # type: ignore
 
     def __repr__(self) -> str:
         'Return the string representation of the RedisDeque.  O(n)'
-        encoded = self.redis.lrange(self.key, 0, -1)
-        values = [self._decode(value) for value in encoded]
+        warnings.warn(
+            cast(str, InefficientAccessWarning.__doc__),
+            InefficientAccessWarning,
+        )
+        encoded_values = self.redis.lrange(self.key, 0, -1)  # Available since Redis 1.0.0
+        values = [self._decode(value) for value in encoded_values]
         repr = self.__class__.__name__ + '(' + str(values)
         if self.maxlen is not None:
             repr += f', maxlen={self.maxlen}'
