@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   nextid.py                                                                 #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -24,32 +24,34 @@ Lua scripting:
 '''
 
 
+# TODO: Remove the following import after deferred evaluation of annotations
+# because the default.
+#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
+#   2. https://www.python.org/dev/peps/pep-0563/
+#   3. https://www.python.org/dev/peps/pep-0649/
+from __future__ import annotations
+
 import concurrent.futures
 import contextlib
-import logging
 from typing import ClassVar
 from typing import Iterable
 from typing import List
-from typing import Optional
 from typing import Tuple
 from typing import Type
 from typing import cast
 
 from redis import Redis
 from redis import RedisError
-from redis.client import Script
-from typing_extensions import Final
+from redis.commands.core import Script
 
 from .base import Primitive
+from .base import logger
 from .exceptions import QuorumIsImpossible
 from .exceptions import QuorumNotAchieved
 from .executor import BailOutExecutor
 
 
-_logger: Final[logging.Logger] = logging.getLogger('pottery')
-
-
-class _Scripts:
+class _Scripts(Primitive):
     '''Parent class to define/register Lua scripts for Redis.
 
     Note that we only have to register these Lua scripts once -- so we do it on
@@ -58,7 +60,7 @@ class _Scripts:
 
     __slots__: Tuple[str, ...] = tuple()
 
-    _set_id_script: ClassVar[Optional[Script]] = None
+    _set_id_script: ClassVar[Script | None] = None
 
     def __init__(self,
                  *,
@@ -66,7 +68,7 @@ class _Scripts:
                  masters: Iterable[Redis] = frozenset(),
                  raise_on_redis_errors: bool = False,
                  ) -> None:
-        super().__init__(  # type: ignore
+        super().__init__(
             key=key,
             masters=masters,
             raise_on_redis_errors=raise_on_redis_errors,
@@ -79,8 +81,9 @@ class _Scripts:
     def __register_set_id_script(self) -> None:
         if self._set_id_script is None:
             class_name = self.__class__.__qualname__
-            _logger.info('Registering %s._set_id_script', class_name)
-            master = next(iter(self.masters))  # type: ignore
+            logger.info('Registering %s._set_id_script', class_name)
+            master = next(iter(self.masters))
+            # Available since Redis 2.6.0:
             self.__class__._set_id_script = master.register_script('''
                 local curr = tonumber(redis.call('get', KEYS[1]))
                 local next = tonumber(ARGV[1])
@@ -93,7 +96,7 @@ class _Scripts:
             ''')
 
 
-class NextId(_Scripts, Primitive):
+class NextId(_Scripts):
     '''Distributed Redis-powered monotonically increasing ID generator.
 
     This algorithm safely and reliably produces monotonically increasing IDs
@@ -134,16 +137,27 @@ class NextId(_Scripts, Primitive):
 
     __slots__ = ('num_tries',)
 
-    KEY_PREFIX: ClassVar[str] = 'nextid'
-    NUM_TRIES: ClassVar[int] = 3
+    _KEY_PREFIX: ClassVar[str] = 'nextid'
+    _NUM_TRIES: ClassVar[int] = 3
 
     def __init__(self,
                  *,
                  key: str = 'current',
                  masters: Iterable[Redis] = frozenset(),
                  raise_on_redis_errors: bool = False,
-                 num_tries: int = NUM_TRIES,
+                 num_tries: int = _NUM_TRIES,
                  ) -> None:
+        '''Initialize a NextId ID generator.
+
+        Keyword arguments:
+            key -- a string that identifies your ID sequence (e.g., 'tweets')
+            masters -- the Redis clients used to achieve quorum for this ID
+                generator
+            raise_on_redis_errors -- whether to raise the QuorumIsImplssible
+                exception when too many Redis masters throw errors
+            num_tries -- the number of times to try to achieve quorum before
+                giving up and raising the QuorumNotAchieved exception
+        '''
         super().__init__(
             key=key,
             masters=masters,
@@ -151,7 +165,7 @@ class NextId(_Scripts, Primitive):
         )
         self.num_tries = num_tries
 
-    def __iter__(self) -> 'NextId':
+    def __iter__(self) -> NextId:
         return self
 
     def __next__(self) -> int:
@@ -176,10 +190,10 @@ class NextId(_Scripts, Primitive):
             current_ids, redis_errors = [], []
             for future in concurrent.futures.as_completed(futures):
                 try:
-                    current_id = int(cast(bytes, future.result() or b'0'))
+                    current_id = int(future.result() or b'0')
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.__current_id() getter caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
@@ -215,7 +229,7 @@ class NextId(_Scripts, Primitive):
                     num_masters_set += future.result() == value
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.__current_id() setter caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
@@ -232,7 +246,7 @@ class NextId(_Scripts, Primitive):
         )
 
     def reset(self) -> None:
-        'Reset the ID counter to 0.'
+        'Reset the ID generator to 0.'
         with concurrent.futures.ThreadPoolExecutor() as executor:
             futures = set()
             for master in self.masters:
@@ -245,7 +259,7 @@ class NextId(_Scripts, Primitive):
                     future.result()
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.reset() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
