@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------- #
 #   redlock.py                                                                #
 #                                                                             #
-#   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
+#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
 #                                                                             #
 #   Licensed under the Apache License, Version 2.0 (the "License");           #
 #   you may not use this file except in compliance with the License.          #
@@ -32,10 +32,16 @@ Lua scripting:
 '''
 
 
+# TODO: Remove the following import after deferred evaluation of annotations
+# because the default.
+#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
+#   2. https://www.python.org/dev/peps/pep-0563/
+#   3. https://www.python.org/dev/peps/pep-0649/
+from __future__ import annotations
+
 import concurrent.futures
 import contextlib
 import functools
-import logging
 import math
 import random
 import time
@@ -45,7 +51,6 @@ from typing import Any
 from typing import Callable
 from typing import ClassVar
 from typing import Iterable
-from typing import Optional
 from typing import Tuple
 from typing import Type
 from typing import cast
@@ -53,12 +58,14 @@ from typing import overload
 
 from redis import Redis
 from redis import RedisError
-from redis.client import Script
-from typing_extensions import Final
+from redis.commands.core import Script
+# TODO: When we drop support for Python 3.7, change the following import to:
+#   from typing import Final
 from typing_extensions import Literal
 
 from .annotations import F
 from .base import Primitive
+from .base import logger
 from .exceptions import ExtendUnlockedLock
 from .exceptions import QuorumNotAchieved
 from .exceptions import ReleaseUnlockedLock
@@ -67,13 +74,7 @@ from .executor import BailOutExecutor
 from .timer import ContextTimer
 
 
-AUTO_RELEASE_TIME: Final[int] = 10 * 1000
-
-
-_logger: Final[logging.Logger] = logging.getLogger('pottery')
-
-
-class _Scripts:
+class _Scripts(Primitive):
     '''Parent class to define/register Lua scripts for Redis.
 
     Note that we only have to register these Lua scripts once -- so we do it on
@@ -82,9 +83,9 @@ class _Scripts:
 
     __slots__: Tuple[str, ...] = tuple()
 
-    _acquired_script: ClassVar[Optional[Script]] = None
-    _extend_script: ClassVar[Optional[Script]] = None
-    _release_script: ClassVar[Optional[Script]] = None
+    _acquired_script: ClassVar[Script | None] = None
+    _extend_script: ClassVar[Script | None] = None
+    _release_script: ClassVar[Script | None] = None
 
     def __init__(self,
                  *,
@@ -92,7 +93,7 @@ class _Scripts:
                  masters: Iterable[Redis] = frozenset(),
                  raise_on_redis_errors: bool = False,
                  ) -> None:
-        super().__init__(  # type: ignore
+        super().__init__(
             key=key,
             masters=masters,
             raise_on_redis_errors=raise_on_redis_errors,
@@ -107,8 +108,9 @@ class _Scripts:
     def __register_acquired_script(self) -> None:
         if self._acquired_script is None:
             class_name = self.__class__.__qualname__
-            _logger.info('Registering %s._acquired_script', class_name)
-            master = next(iter(self.masters))  # type: ignore
+            logger.info('Registering %s._acquired_script', class_name)
+            master = next(iter(self.masters))
+            # Available since Redis 2.6.0:
             self.__class__._acquired_script = master.register_script('''
                 if redis.call('get', KEYS[1]) == ARGV[1] then
                     local pttl = redis.call('pttl', KEYS[1])
@@ -121,8 +123,9 @@ class _Scripts:
     def __register_extend_script(self) -> None:
         if self._extend_script is None:
             class_name = self.__class__.__qualname__
-            _logger.info('Registering %s._extend_script', class_name)
-            master = next(iter(self.masters))  # type: ignore
+            logger.info('Registering %s._extend_script', class_name)
+            master = next(iter(self.masters))
+            # Available since Redis 2.6.0:
             self.__class__._extend_script = master.register_script('''
                 if redis.call('get', KEYS[1]) == ARGV[1] then
                     return redis.call('pexpire', KEYS[1], ARGV[2])
@@ -134,8 +137,9 @@ class _Scripts:
     def __register_release_script(self) -> None:
         if self._release_script is None:
             class_name = self.__class__.__qualname__
-            _logger.info('Registering %s._release_script', class_name)
-            master = next(iter(self.masters))  # type: ignore
+            logger.info('Registering %s._release_script', class_name)
+            master = next(iter(self.masters))
+            # Available since Redis 2.6.0:
             self.__class__._release_script = master.register_script('''
                 if redis.call('get', KEYS[1]) == ARGV[1] then
                     return redis.call('del', KEYS[1])
@@ -145,7 +149,7 @@ class _Scripts:
             ''')
 
 
-class Redlock(_Scripts, Primitive):
+class Redlock(_Scripts):
     '''Distributed Redis-powered lock.
 
     This algorithm safely and reliably provides a mutually-exclusive locking
@@ -162,11 +166,10 @@ class Redlock(_Scripts, Primitive):
         >>> printer_lock = Redlock(key='printer', masters={redis})
         >>> bool(printer_lock.locked())
         False
-        >>> printer_lock.acquire()
-        True
-        >>> bool(printer_lock.locked())
-        True
-        >>> # Critical section - print stuff here.
+        >>> if printer_lock.acquire():
+        ...     print('printer_lock is locked')
+        ...     # Critical section - print stuff here.
+        printer_lock is locked
         >>> printer_lock.release()
         >>> bool(printer_lock.locked())
         False
@@ -178,11 +181,9 @@ class Redlock(_Scripts, Primitive):
     (in the event that a process dies inside a critical section before it
     releases its lock).
 
-        >>> printer_lock.acquire()
-        True
-        >>> bool(printer_lock.locked())
-        True
-        >>> # Critical section - print stuff here.
+        >>> if printer_lock.acquire():
+        ...     # Critical section - print stuff here.
+        ...     time.sleep(10)
         >>> time.sleep(10)
         >>> bool(printer_lock.locked())
         False
@@ -190,13 +191,10 @@ class Redlock(_Scripts, Primitive):
     If 10 seconds isn't enough to complete executing your critical section,
     then you can specify your own timeout:
 
-        >>> printer_lock = Redlock(key='printer', masters={redis}, auto_release_time=15*1000)
-        >>> printer_lock.acquire()
-        True
-        >>> bool(printer_lock.locked())
-        True
-        >>> # Critical section - print stuff here.
-        >>> time.sleep(10)
+        >>> printer_lock = Redlock(key='printer', masters={redis}, auto_release_time=15)
+        >>> if printer_lock.acquire():
+        ...     # Critical section - print stuff here.
+        ...     time.sleep(10)
         >>> bool(printer_lock.locked())
         True
         >>> time.sleep(5)
@@ -205,21 +203,19 @@ class Redlock(_Scripts, Primitive):
 
     You can use a Redlock as a context manager:
 
-        >>> states = []
         >>> with Redlock(key='printer', masters={redis}) as printer_lock:
-        ...     states.append(bool(printer_lock.locked()))
+        ...     print('printer_lock is locked')
         ...     # Critical section - print stuff here.
-        >>> states.append(bool(printer_lock.locked()))
-        >>> states
-        [True, False]
+        printer_lock is locked
+        >>> bool(printer_lock.locked())
+        False
 
-        >>> states = []
         >>> with printer_lock:
-        ...     states.append(bool(printer_lock.locked()))
+        ...     print('printer_lock is locked')
         ...     # Critical section - print stuff here.
-        >>> states.append(bool(printer_lock.locked()))
-        >>> states
-        [True, False]
+        printer_lock is locked
+        >>> bool(printer_lock.locked())
+        False
     '''
 
     __slots__ = (
@@ -231,21 +227,41 @@ class Redlock(_Scripts, Primitive):
         '_extension_num',
     )
 
-    KEY_PREFIX: ClassVar[str] = 'redlock'
-    CLOCK_DRIFT_FACTOR: ClassVar[float] = 0.01
-    RETRY_DELAY: ClassVar[int] = 200
-    NUM_EXTENSIONS: ClassVar[int] = 3
+    _KEY_PREFIX: ClassVar[str] = 'redlock'
+    _AUTO_RELEASE_TIME: ClassVar[float] = 10
+    _CLOCK_DRIFT_FACTOR: ClassVar[float] = 0.01
+    _RETRY_DELAY: ClassVar[float] = .2
+    _NUM_EXTENSIONS: ClassVar[int] = 3
 
     def __init__(self,
                  *,
                  key: str,
                  masters: Iterable[Redis] = frozenset(),
                  raise_on_redis_errors: bool = False,
-                 auto_release_time: int = AUTO_RELEASE_TIME,
-                 num_extensions: int = NUM_EXTENSIONS,
+                 auto_release_time: float = _AUTO_RELEASE_TIME,
+                 num_extensions: int = _NUM_EXTENSIONS,
                  context_manager_blocking: bool = True,
                  context_manager_timeout: float = -1,
                  ) -> None:
+        '''Initialize a Redlock.
+
+        Keyword arguments:
+            key -- a string that identifies your resource
+            masters -- the Redis clients used to achieve quorum for this
+                Redlock's state
+            raise_on_redis_errors -- whether to raise the QuorumIsImplssible
+                exception when too many Redis masters throw errors
+            auto_release_time -- the timeout in seconds by which to
+                automatically release this Redlock, unless it's already been
+                released
+            num_extensions -- the number of times that this Redlock's lease can
+                be extended
+            context_manager_blocking -- when using this Redlock as a context
+                manager, whether to block when acquiring
+            context_manager_timeout -- if context_manager_blocking, how long to
+                wait when acquiring before giving up and raising the
+                QuorumNotAchieved exception
+        '''
         if not context_manager_blocking and context_manager_timeout != -1:
             raise ValueError("can't specify a timeout for a non-blocking call")
 
@@ -265,7 +281,7 @@ class Redlock(_Scripts, Primitive):
         acquired = master.set(
             self.key,
             self._uuid,
-            px=self.auto_release_time,
+            px=int(self.auto_release_time * 1000),
             nx=True,
         )
         return bool(acquired)
@@ -282,9 +298,10 @@ class Redlock(_Scripts, Primitive):
         return ttl
 
     def __extend_master(self, master: Redis) -> bool:
+        auto_release_time_ms = int(self.auto_release_time * 1000)
         extended = cast(Script, self._extend_script)(
             keys=(self.key,),
-            args=(self._uuid, self.auto_release_time),
+            args=(self._uuid, auto_release_time_ms),
             client=master,
         )
         return bool(extended)
@@ -298,11 +315,11 @@ class Redlock(_Scripts, Primitive):
         return bool(released)
 
     def __drift(self) -> float:
-        return self.auto_release_time * self.CLOCK_DRIFT_FACTOR + 2
+        return self.auto_release_time * self._CLOCK_DRIFT_FACTOR + .002
 
     def __acquire_masters(self,
                           *,
-                          raise_on_redis_errors: Optional[bool] = None,
+                          raise_on_redis_errors: bool | None = None,
                           ) -> bool:
         self._uuid = str(uuid.uuid4())
         self._extension_num = 0
@@ -319,7 +336,7 @@ class Redlock(_Scripts, Primitive):
                     num_masters_acquired += future.result()
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.__acquire_masters() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
@@ -327,8 +344,8 @@ class Redlock(_Scripts, Primitive):
                 else:
                     if num_masters_acquired > len(self.masters) // 2:
                         validity_time = self.auto_release_time
-                        validity_time -= round(self.__drift())
-                        validity_time -= timer.elapsed()
+                        validity_time -= self.__drift()
+                        validity_time -= timer.elapsed() / 1000
                         if validity_time > 0:  # pragma: no cover
                             return True
 
@@ -341,7 +358,7 @@ class Redlock(_Scripts, Primitive):
                 *,
                 blocking: bool = True,
                 timeout: float = -1,
-                raise_on_redis_errors: Optional[bool] = None,
+                raise_on_redis_errors: bool | None = None,
                 ) -> bool:
         '''Lock the lock.
 
@@ -396,7 +413,7 @@ class Redlock(_Scripts, Primitive):
         def log_time_enqueued(timer: ContextTimer, acquired: bool) -> None:
             key_suffix = self.key.split(':', maxsplit=1)[1]
             time_enqueued = math.ceil(timer.elapsed())
-            _logger.info(
+            logger.info(
                 'source=pottery sample#redlock.enqueued.%s=%dms sample#redlock.acquired.%s=%d',
                 key_suffix,
                 time_enqueued,
@@ -413,7 +430,8 @@ class Redlock(_Scripts, Primitive):
                             log_time_enqueued(timer, True)
                         return True
                     enqueued = True
-                    time.sleep(random.uniform(0, self.RETRY_DELAY/1000))
+                    delay = random.uniform(0, self._RETRY_DELAY)  # nosec
+                    time.sleep(delay)
             if enqueued:
                 log_time_enqueued(timer, False)
             return False
@@ -425,7 +443,7 @@ class Redlock(_Scripts, Primitive):
 
     __acquire = acquire
 
-    def locked(self, *, raise_on_redis_errors: Optional[bool] = None) -> int:
+    def locked(self, *, raise_on_redis_errors: bool | None = None) -> float:
         '''How much longer we'll hold the lock (unless we extend or release it).
 
         If we don't currently hold the lock, then this method returns 0.
@@ -448,7 +466,7 @@ class Redlock(_Scripts, Primitive):
 
             >>> printer_lock_1.acquire()
             True
-            >>> 9 * 1000 < printer_lock_1.locked() < 10 * 1000
+            >>> 9 < printer_lock_1.locked() < 10
             True
             >>> printer_lock_1.release()
 
@@ -465,26 +483,24 @@ class Redlock(_Scripts, Primitive):
                     ttl = future.result()
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.locked() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
                     )
                 else:
                     if ttl:
-                        ttls.append(ttl)
+                        ttls.append(ttl / 1000)
                         if len(ttls) > len(self.masters) // 2:  # pragma: no cover
                             validity_time = min(ttls)
-                            validity_time -= round(self.__drift())
-                            validity_time -= timer.elapsed()
+                            validity_time -= self.__drift()
+                            validity_time -= timer.elapsed() / 1000
                             return max(validity_time, 0)
 
         self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
         return 0
 
-    __locked = locked
-
-    def extend(self, *, raise_on_redis_errors: Optional[bool] = None) -> None:
+    def extend(self, *, raise_on_redis_errors: bool | None = None) -> None:
         '''Extend our hold on the lock (if we currently hold it).
 
         Usage:
@@ -494,13 +510,13 @@ class Redlock(_Scripts, Primitive):
             >>> printer_lock = Redlock(key='printer', masters={redis})
             >>> printer_lock.acquire()
             True
-            >>> 9 * 1000 < printer_lock.locked() < 10 * 1000
+            >>> 9 < printer_lock.locked() < 10
             True
             >>> time.sleep(1)
-            >>> 8 * 1000 < printer_lock.locked() < 9 * 1000
+            >>> 8 < printer_lock.locked() < 9
             True
             >>> printer_lock.extend()
-            >>> 9 * 1000 < printer_lock.locked() < 10 * 1000
+            >>> 9 < printer_lock.locked() < 10
             True
             >>> printer_lock.release()
         '''
@@ -519,7 +535,7 @@ class Redlock(_Scripts, Primitive):
                     num_masters_extended += future.result()
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.extend() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
@@ -536,7 +552,7 @@ class Redlock(_Scripts, Primitive):
             redis_errors=redis_errors,
         )
 
-    def release(self, *, raise_on_redis_errors: Optional[bool] = None) -> None:
+    def release(self, *, raise_on_redis_errors: bool | None = None) -> None:
         '''Unlock the lock.
 
         Usage:
@@ -566,7 +582,7 @@ class Redlock(_Scripts, Primitive):
                     num_masters_released += future.result()
                 except RedisError as error:
                     redis_errors.append(error)
-                    _logger.exception(
+                    logger.exception(
                         '%s.release() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
@@ -584,28 +600,26 @@ class Redlock(_Scripts, Primitive):
 
     __release = release
 
-    def __enter__(self) -> 'Redlock':
+    def __enter__(self) -> Redlock:
         '''You can use a Redlock as a context manager.
 
         Usage:
 
-            >>> states = []
             >>> from redis import Redis
             >>> redis = Redis(socket_timeout=1)
             >>> with Redlock(key='printer', masters={redis}) as printer_lock:
-            ...     states.append(bool(printer_lock.locked()))
+            ...     print('printer_lock is locked')
             ...     # Critical section - print stuff here.
-            >>> states.append(bool(printer_lock.locked()))
-            >>> states
-            [True, False]
+            printer_lock is locked
+            >>> bool(printer_lock.locked())
+            False
 
-            >>> states = []
             >>> with printer_lock:
-            ...     states.append(bool(printer_lock.locked()))
+            ...     print('printer_lock is locked')
             ...     # Critical section - print stuff here.
-            >>> states.append(bool(printer_lock.locked()))
-            >>> states
-            [True, False]
+            printer_lock is locked
+            >>> bool(printer_lock.locked())
+            False
         '''
         acquired = self.__acquire(
             blocking=self.context_manager_blocking,
@@ -632,31 +646,29 @@ class Redlock(_Scripts, Primitive):
         raise NotImplementedError
 
     def __exit__(self,
-                 exc_type: Optional[Type[BaseException]],
-                 exc_value: Optional[BaseException],
-                 traceback: Optional[TracebackType],
+                 exc_type: Type[BaseException] | None,
+                 exc_value: BaseException | None,
+                 traceback: TracebackType | None,
                  ) -> Literal[False]:
         '''You can use a Redlock as a context manager.
 
         Usage:
 
-            >>> states = []
             >>> from redis import Redis
             >>> redis = Redis(socket_timeout=1)
             >>> with Redlock(key='printer', masters={redis}) as printer_lock:
-            ...     states.append(bool(printer_lock.locked()))
+            ...     print('printer_lock is locked')
             ...     # Critical section - print stuff here.
-            >>> states.append(bool(printer_lock.locked()))
-            >>> states
-            [True, False]
+            printer_lock is locked
+            >>> bool(printer_lock.locked())
+            False
 
-            >>> states = []
             >>> with printer_lock:
-            ...     states.append(bool(printer_lock.locked()))
+            ...     print('printer_lock is locked')
             ...     # Critical section - print stuff here.
-            >>> states.append(bool(printer_lock.locked()))
-            >>> states
-            [True, False]
+            printer_lock is locked
+            >>> bool(printer_lock.locked())
+            False
         '''
         self.__release()
         return False
@@ -669,7 +681,7 @@ def synchronize(*,
                 key: str,
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
-                auto_release_time: int = AUTO_RELEASE_TIME,
+                auto_release_time: float = Redlock._AUTO_RELEASE_TIME,
                 blocking: bool = True,
                 timeout: float = -1,
                 ) -> Callable[[F], F]:
@@ -679,9 +691,25 @@ def synchronize(*,
     function at a time.  Under the hood, synchronize() uses a Redlock.  See
     help(Redlock) for more details.
 
+    Keyword arguments:
+        key -- a string that identifies your resource
+        masters -- the Redis clients used to achieve quorum for this Redlock's
+            state
+        raise_on_redis_errors -- whether to raise the QuorumIsImplssible
+            exception when too many Redis masters throw errors
+        auto_release_time -- the timeout in seconds by which to automatically
+            release this Redlock, unless it's already been released
+        num_extensions -- the number of times that this Redlock's lease can be
+            extended
+        context_manager_blocking -- when using this Redlock as a context
+            manager, whether to block when acquiring
+        context_manager_timeout -- if context_manager_blocking, how long to wait
+            when acquiring before giving up and raising the QuorumNotAchieved
+            exception
+
     Usage:
 
-        >>> @synchronize(key='synchronized-func', auto_release_time=1500)
+        >>> @synchronize(key='synchronized-func', auto_release_time=1.5)
         ... def func():
         ...     # Only one thread can execute this function at a time.
         ...     return True
@@ -726,7 +754,7 @@ def _log_synchronize(func: F,
                      holding_timer: ContextTimer,
                      ) -> None:
     try:
-        _logger.info(
+        logger.info(
             '%s() waited for %s for %d ms; held for %d ms',
             func.__qualname__,
             redlock.key,
@@ -737,7 +765,7 @@ def _log_synchronize(func: F,
         # holding_timer.elapsed() threw a RuntimeError, which means that
         # holding_timer never started, which means that we never acquired the
         # lock / entered the critical section.
-        _logger.info(
+        logger.info(
             '%s() waited for %s for %d ms; never acquired lock',
             func.__qualname__,
             redlock.key,
